#include "FakeSurface.h"

#include "BitmapSaver.h"

FakeSurface::FakeSurface(IDirectDraw4* ddraw) {
	Refs=1;
	MenuMode=true;
	memset(&rDesc, 0, sizeof(DDSURFACEDESC2));
	rDesc.dwSize=sizeof(DDSURFACEDESC2);
	rDesc.dwFlags=DDSD_CAPS|DDSD_BACKBUFFERCOUNT;
	rDesc.ddsCaps.dwCaps=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX|DDSCAPS_3DDEVICE;
	rDesc.dwBackBufferCount=1;

	HRESULT hr=ddraw->CreateSurface(&rDesc, &Real, 0);
	CHECKFAIL(hr);
	hr=Real->GetSurfaceDesc(&rDesc);
	CHECKFAIL(hr);

	DDSCAPS2 caps;
	memset(&caps, 0, sizeof(DDSCAPS2));
	caps.dwCaps=DDSCAPS_BACKBUFFER;
	hr=Real->GetAttachedSurface(&caps, &Real2);
	CHECKFAIL(hr);

	rDesc.ddpfPixelFormat.dwRBitMask=0xf800;
	rDesc.ddpfPixelFormat.dwGBitMask=0x7e0;
	rDesc.ddpfPixelFormat.dwBBitMask=0x1f;
	rDesc.ddpfPixelFormat.dwRGBBitCount=16;

	memset(&mDesc, 0, sizeof(DDSURFACEDESC2));
	mDesc.dwSize=sizeof(DDSURFACEDESC2);
	mDesc.dwFlags=DDSD_CAPS|DDSD_PIXELFORMAT|DDSD_HEIGHT|DDSD_WIDTH;
	if(UseSysMemOverlay==1) {
		mDesc.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_SYSTEMMEMORY;
	} else {
		mDesc.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_VIDEOMEMORY;
	}
	mDesc.ddpfPixelFormat.dwSize=sizeof(DDPIXELFORMAT);
	mDesc.ddpfPixelFormat.dwFlags=DDPF_RGB;
	mDesc.ddpfPixelFormat.dwRBitMask=0xf800;
	mDesc.ddpfPixelFormat.dwGBitMask=0x7e0;
	mDesc.ddpfPixelFormat.dwBBitMask=0x1f;
	mDesc.ddpfPixelFormat.dwRGBBitCount=16;
	mDesc.dwWidth=640;
	mDesc.dwHeight=480;

	oDesc=mDesc;
	oDesc.dwWidth=gWidth;
	oDesc.dwHeight=gHeight;
	if(UseSysMemOverlay==2) oDesc.ddsCaps.dwCaps=DDSCAPS_OFFSCREENPLAIN|DDSCAPS_SYSTEMMEMORY;

	hr=ddraw->CreateSurface(&mDesc, &menu, 0);
	CHECKFAIL(hr);
	hr=menu->GetSurfaceDesc(&mDesc);
	CHECKFAIL(hr);
	mDesc.ddsCaps.dwCaps|=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX;
	mDesc.dwBackBufferCount=1;
	mDesc.dwFlags|=DDSD_BACKBUFFERCOUNT;

	hr=ddraw->CreateSurface(&oDesc, &overlay, 0);
	CHECKFAIL(hr);
	hr=overlay->GetSurfaceDesc(&oDesc);
	CHECKFAIL(hr);
	if(UseSysMemOverlay!=2) {
		DDCOLORKEY key;
		key.dwColorSpaceHighValue=TextOverlayColourKey;
		key.dwColorSpaceLowValue=TextOverlayColourKey;
		hr=overlay->SetColorKey(DDCKEY_SRCBLT, &key);
		CHECKFAIL(hr);
	}
	oDesc.ddsCaps.dwCaps|=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX;
	oDesc.dwBackBufferCount=1;
	oDesc.dwFlags|=DDSD_BACKBUFFERCOUNT;

	ClearOverlay();
}
void FakeSurface::ClearOverlay() {
	DDBLTFX fx;
	memset(&fx, 0, sizeof(DDBLTFX));
	fx.dwSize=sizeof(DDBLTFX);
	fx.dwFillColor=TextOverlayColourKey;
	HRESULT hr=overlay->Blt(0, 0, 0, DDBLT_WAIT|DDBLT_COLORFILL, &fx);
	CHECKFAIL(hr);
}
void FakeSurface::SetMenuMode(bool b) { MenuMode=b; GetAsyncKeyState(VK_SNAPSHOT); }
IDirectDrawSurface4* FakeSurface::QueryReal() { return Real2; }
// IUnknown methods
HRESULT _stdcall FakeSurface::QueryInterface(REFIID a, LPVOID * b) {
	//69c11c3e = IID_IDirectDrawGammaControl
	//93281502 = IID_IDirect3DTexture2
	if(a==IID_IDirectDrawGammaControl) {
		return Real->QueryInterface(a,b);
	} else UNUSEDFUNCTION;
}
ULONG _stdcall FakeSurface::AddRef()  { return ++Refs; }
ULONG _stdcall FakeSurface::Release() {
	if(!--Refs) {
		menu->Release();
		Real->Release();
		delete this;
		return 0;
	} else return Refs;
}
// IDirectDrawSurface methods
HRESULT _stdcall FakeSurface::AddAttachedSurface(LPDIRECTDRAWSURFACE4 a) {
	return Real2->AddAttachedSurface(a);
}
HRESULT _stdcall FakeSurface::AddOverlayDirtyRect(LPRECT) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::Blt(LPRECT,LPDIRECTDRAWSURFACE4, LPRECT,DWORD, LPDDBLTFX) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::BltBatch(LPDDBLTBATCH, DWORD, DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::BltFast(DWORD,DWORD,LPDIRECTDRAWSURFACE4, LPRECT,DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::DeleteAttachedSurface(DWORD a,LPDIRECTDRAWSURFACE4 b) {
	return Real2->DeleteAttachedSurface(a,b);
}
HRESULT _stdcall FakeSurface::EnumAttachedSurfaces(LPVOID,LPDDENUMSURFACESCALLBACK2) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::EnumOverlayZOrders(DWORD,LPVOID,LPDDENUMSURFACESCALLBACK2) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::Flip(LPDIRECTDRAWSURFACE4 a, DWORD b) {
#define BITS16TO32(a) ( (((a)&0xf800)<<8) + (((a)&0x7e0)<<5) + (((a)&0x1f)<<3) )
	if(MenuMode) UNUSEDFUNCTION;
	HRESULT hr;
	if(UseSysMemOverlay==2) {
		DDSURFACEDESC2 realdesc;
		DDSURFACEDESC2 overdesc;
		realdesc.dwSize=overdesc.dwSize=sizeof(DDSURFACEDESC2);
		hr=overlay->Lock(0, &overdesc, DDLOCK_READONLY|DDLOCK_WAIT, 0);
		hr=Real2->Lock(0, &realdesc, DDLOCK_WAIT, 0);
		realdesc.lPitch/=4;
		overdesc.lPitch/=2;
		for(DWORD y=0;y<gHeight;y++) {
			for(DWORD x=0;x<gWidth;x++) {
				if(((WORD*)overdesc.lpSurface)[y*overdesc.lPitch+x]!=TextOverlayColourKey)
					((DWORD*)realdesc.lpSurface)[y*realdesc.lPitch+x]=BITS16TO32(((WORD*)overdesc.lpSurface)[y*overdesc.lPitch+x]);
			}
		}
		hr=Real2->Unlock(0);
		hr=overlay->Unlock(0);
	} else {
		hr=Real2->BltFast(0, 0, overlay, 0, DDBLTFAST_SRCCOLORKEY|DDBLTFAST_WAIT);
		CHECKFAIL(hr);
	}
	if(GetAsyncKeyState(VK_SNAPSHOT)) {
		DDSURFACEDESC2 desc;
		memset(&desc, 0, sizeof(desc));
		desc.dwSize=sizeof(desc);
		Real2->Lock(0, &desc, DDLOCK_READONLY|DDLOCK_WAIT, 0);
		SaveBitmap(desc.lpSurface, desc.lPitch);
		Real2->Unlock(0);
	}
	ClearOverlay();
	hr=Real->Flip(a,b);
	CHECKFAIL(hr);
	return DD_OK;
}
HRESULT _stdcall FakeSurface::GetAttachedSurface(LPDDSCAPS2 a, LPDIRECTDRAWSURFACE4 * b) {
	if(a->dwCaps&DDSCAPS_BACKBUFFER) {
		AddRef();
		*b=this;
		return DD_OK;
	} else UNUSEDFUNCTION;
}
HRESULT _stdcall FakeSurface::GetBltStatus(DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetCaps(LPDDSCAPS2 a) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetClipper(LPDIRECTDRAWCLIPPER *) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetColorKey(DWORD, LPDDCOLORKEY) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetDC(HDC *) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetFlipStatus(DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetOverlayPosition(LPLONG, LPLONG) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetPalette(LPDIRECTDRAWPALETTE *) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetPixelFormat(LPDDPIXELFORMAT) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetSurfaceDesc(LPDDSURFACEDESC2 a) {
	*a=MenuMode?mDesc:oDesc;
	return DD_OK;
}
HRESULT _stdcall FakeSurface::Initialize(LPDIRECTDRAW, LPDDSURFACEDESC2) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::IsLost() {
	return Real->IsLost();
}
HRESULT _stdcall FakeSurface::Lock(LPRECT a,LPDDSURFACEDESC2 b,DWORD c,HANDLE d) {
	if(MenuMode) {
		HRESULT hr=menu->Lock(a,b,c,d);
		CHECKFAIL(hr);
		if(FAILED(hr)) return hr;
		b->ddsCaps.dwCaps|=DDSCAPS_VIDEOMEMORY|DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX;
		b->dwBackBufferCount=1;
		b->dwFlags|=DDSD_BACKBUFFERCOUNT;
		return DD_OK;
	} else {
		HRESULT hr=overlay->Lock(a,b,c,d);
		CHECKFAIL(hr);
		if(FAILED(hr)) return hr;
		return DD_OK;
	}
}
HRESULT _stdcall FakeSurface::ReleaseDC(HDC) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::Restore() {
	switch(UseSysMemOverlay) {
		case 0: overlay->Restore();
		case 2: menu->Restore();
	}
	return Real->Restore();
}
HRESULT _stdcall FakeSurface::SetClipper(LPDIRECTDRAWCLIPPER) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::SetColorKey(DWORD, LPDDCOLORKEY) { UNUSEDFUNCTION;	}
HRESULT _stdcall FakeSurface::SetOverlayPosition(LONG, LONG) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::SetPalette(LPDIRECTDRAWPALETTE) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::Unlock(LPRECT a) {
	if(MenuMode) {
		HRESULT hr=menu->Unlock(a);
		CHECKFAIL(hr);
		if(FAILED(hr)) return hr;
		hr=Real->Blt(0, menu, 0, 0, 0);
		return DD_OK;
	} else {
		overlay->Unlock(a);
		return DD_OK;
	}
}
HRESULT _stdcall FakeSurface::UpdateOverlay(LPRECT, LPDIRECTDRAWSURFACE4,LPRECT,DWORD, LPDDOVERLAYFX) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::UpdateOverlayDisplay(DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::UpdateOverlayZOrder(DWORD, LPDIRECTDRAWSURFACE4) { UNUSEDFUNCTION; }
//*** Added in the v2 interface ***
HRESULT _stdcall FakeSurface::GetDDInterface(LPVOID*) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::PageLock(DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::PageUnlock(DWORD) { UNUSEDFUNCTION; }
//*** Added in the v3 interface ***
HRESULT _stdcall FakeSurface::SetSurfaceDesc(LPDDSURFACEDESC2, DWORD) { UNUSEDFUNCTION; }
//*** Added in the v4 interface ***
HRESULT _stdcall FakeSurface::SetPrivateData(REFGUID, LPVOID, DWORD, DWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetPrivateData(REFGUID, LPVOID, LPDWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::FreePrivateData(REFGUID) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::GetUniquenessValue(LPDWORD) { UNUSEDFUNCTION; }
HRESULT _stdcall FakeSurface::ChangeUniquenessValue() { UNUSEDFUNCTION; }